#include <math.h>
#include <mex.h>

bool Display = false;

#include "chili_liouvmatrix_func.cpp"

const long maxElements = 500000;
const int maxRows = 20000;

const double sqrt12 = sqrt(1.0/2.0);
const double sqrt13 = sqrt(1.0/3.0);
const double sqrt23 = sqrt(2.0/3.0);

double *ridx, *cidx;
double *MatrixRe, *MatrixIm;

int nRows, nElements;

struct SystemStruct {
  double I;
  double EZI0, NZI0, HFI0;
  double *ReEZI2, *ImEZI2, *ReHFI2, *ImHFI2;
};

struct BasisStruct {
  int evenLmax, oddLmax, Kmax, Mmax;
  int jKmin, pSmin, pImax, deltaL, deltaK;
};

struct DiffusionStruct {
  double* xlk, *d2psi;
  double psi, Rxx, Ryy, Rzz;
  double Exchange, JumpFreq;
  int nJumpSites;
};

SystemStruct Sys;
BasisStruct Basis;
DiffusionStruct Diff;

//============================================================================
//============================================================================
//============================================================================
//============================================================================
void makematrix()
{

const double I = Sys.I;
const double NZI0 = Sys.NZI0;
const double HFI0 = Sys.HFI0;
const double EZI0 = Sys.EZI0;
const double *ReEZI2 = Sys.ReEZI2;
const double *ImEZI2 = Sys.ImEZI2;
const double *ReHFI2 = Sys.ReHFI2;
const double *ImHFI2 = Sys.ImHFI2;

// Diffusion parameters
//---------------------------------------------
const double psi = Diff.psi;
const double *d2psi = Diff.d2psi;
const double Rxx = Diff.Rxx;
const double Ryy = Diff.Ryy;
const double Rzz = Diff.Rzz;
const double Rxy2 = 0; // anisotropic viscosity
const double Rzz2 = 0;

const int nJumpSites = Diff.nJumpSites; // number of jump sites
const double JumpFreq = Diff.JumpFreq;

const double pl = 0, pmxl = 0, pmz = 0, pkzz = 0, pmxk = 0, pkxy = 0;

const int ipt = 0; // number of potential terms
const double ExchangeFreq = Diff.Exchange;
const double *xlk = Diff.xlk;

// Basis parameters
//---------------------------------------------
const int Lemax = Basis.evenLmax;
const int Lomax = Basis.oddLmax;
const int Kmax = Basis.Kmax;
const int Mmax = Basis.Mmax;
const int jKmin = Basis.jKmin;
const int pSmin = Basis.pSmin;
const int pImax = Basis.pImax;
const int deltaL = Basis.deltaL;
const int deltaK = Basis.deltaK;
//-------------------------------------------------

// Lband = (lptmx>=2) ? lptmx*2 : 2;
const int Lband = 8;
// Kband = kptmx*2;
const int Kband = 8;
//==========================================================

// Loop variables
int L1, jK1, K1, M1, pS1, qS1, pI1, qI1;
int L2, jK2, K2, M2, pS2, qS2, pI2, qI2;
int Ld, Ls, jKd, Kd, Ks;

// Min and max values for loop variables
int K1max, M1max, qS1max, pI1max, qI1max;
int L2max, jK2min, K2min, K2max, M2min, M2max;
int pS2min, qS2min, qS2max, pI2min, pI2max, qI2min, qI2max;

int iRow = 0, iCol = 0, iElement = 0;

// Brownian model if all three exponents in isotropic
// diffusion term are zero.
bool Brownian = (pl==0) && (pkxy==0) && (pkzz==0);
bool AnisoVisc = (Rxy2!=0) || (Rzz2!=0);
bool DiscreteJumps = (nJumpSites!=0);
bool Potential = (ipt>0);
bool Exchange = (ExchangeFreq!=0);

/*
mexPrintf("   Brownian:  %d\n",Brownian);
mexPrintf("   AnisoVisc: %d\n",AnisoVisc);
mexPrintf("   Potential: %d\n",Potential);
mexPrintf("   Exchange:  %d\n",Exchange);
mexPrintf("   DiscJumps: %d\n",DiscreteJumps);
mexPrintf("   Lemax Lomax deltaL:   %d  %d  %d\n",Lemax,Lomax,deltaL);
mexPrintf("   jKmin Kmax Mmax:   %d  %d  %d\n",jKmin,Kmax,Mmax);
mexPrintf("   pSmin  pImax:   %d  %d\n",pSmin, pImax);
*/

double IsoDiffDiagLK, IsoDiffLKm2, IsoDiffLKp2, IsoDiffDiag, PotDiff;

const bool AxialDiff = (Rxx==Ryy);

for (L1=0;L1<=Lemax;L1+=deltaL) {
  if (isodd(L1) && (L1>Lomax)) continue;
  for (jK1=jKmin;jK1<=1;jK1+=2) {
    K1max = mini(Kmax,L1);
    for (K1=0;K1<=K1max;K1+=deltaK) {
      if ((K1==0)&&(parity(L1)!=jK1)) continue;

      // Isotropic part of diffusion operator
      //--------------------------------------------------------
      if (Brownian) {
        IsoDiffDiagLK = (Rxx+Ryy)/2*(L1*(L1+1)) + K1*K1*(Rzz-(Rxx+Ryy)/2);
        if (!AxialDiff) {
          int KK = K1-2;
          IsoDiffLKm2 = (Rxx-Ryy)/4*sqrt((L1-KK-1)*(L1-KK)*(L1+KK+1)*(L1+KK+2));
          KK = K1+2;
          IsoDiffLKp2 = (Rxx-Ryy)/4*sqrt((L1+KK-1)*(L1+KK)*(L1-KK+1)*(L1-KK+2));
        }
      }
      else {
        IsoDiffDiagLK = Rxx*L1*(L1+1)*pow(1+pmxl*L1*(L1+1),-pl) +
                        K1*K1*(Rzz*pow(1+ pmz*K1*K1,-pkzz) -
                               Rxx*pow(1+pmxk*K1*K1,-pkxy));
      }
      //--------------------------------------------------------
      
      // Discrete jump motion
      //--------------------------------------------------------
      if (DiscreteJumps)
        if ((K1%nJumpSites)!=0) IsoDiffDiagLK += JumpFreq;
      //--------------------------------------------------------
      
      M1max = mini(Mmax,L1);
      for (M1=-M1max;M1<=M1max;M1++) {

        // Anisotropic viscosity term in diffusion operator
        //--------------------------------------------------
        IsoDiffDiag = IsoDiffDiagLK;
        if (AnisoVisc) IsoDiffDiag += M1*M1*(Rzz2-Rxy2);
        //--------------------------------------------------
        
        for (pS1=pSmin;pS1<=1;pS1++) {
          qS1max = 1 - abs(pS1);
          for (qS1=-qS1max;qS1<=qS1max;qS1+=2) {
            for (pI1 = -pImax;pI1<=pImax;pI1++) {
              if (psi==0) if ((pI1+pS1-1)!=M1) continue;
              qI1max = int(2*I) - abs(pI1);
              for (qI1=-qI1max;qI1<=qI1max;qI1+=2) {
                //mexPrintf("%3d %3d %3d %3d    %2d %2d %2d %2d\n",L1,jK1,K1,M1,pS1,qS1,pI1,qI1);
                iCol = iRow;
                bool diagRC = true;
                L2max = mini(Lemax,L1+Lband);
                for (L2=L1;L2<=L2max;L2+=deltaL) {
                  if (isodd(L2)&&(L2>Lomax)) continue;
                  Ld = L1 - L2; Ls = L1 + L2;
                  bool Ld2 = abs(Ld)<=2;

                  // N_L normalisation factor, see after Eq. (A11)
                  const double N_L = sqrt(double((2*L1+1)*(2*L2+1)));

                  jK2min  = (diagRC) ?  jK1 : jKmin;
                  for (jK2=jK2min;jK2<=1;jK2+=2) {
                    jKd = jK1 - jK2;
                    K2max = mini(Kmax,L2);
                    K2min = (diagRC) ? K1 : 0;
                    for (K2=K2min;K2<=K2max;K2+=deltaK) {
                      if ((K2==0)&&(parity(L2)!=jK2)) continue;
                      bool diagLK = (L1==L2) && (K1==K2);
                      Kd = K1 - K2;
                      Ks = K1 + K2;
                      int parityLK2 = parity(L2+K2);
                      
                      //----------------------------------------------
                      // R(mu=EZI,HFI;l=2), see Eq. (A42) and (A44)
                      //----------------------------------------------
                      double R_EZI2 = 0, R_HFI2 = 0;
                      if (Ld2) {
                        double a1 = 0, g1 = 0;
                        if (abs(Kd)<=2) {
                          const double coeff = jjj(L1,2,L2,K1,-Kd,-K2);
                          //mexPrintf("[wigner3j(%d,%d,%d,%d,%d,%d) %g]\n",L1,2,L2,K1,-Kd,-K2,coeff);
                          if (jK1==jK2) {
                            a1 = coeff*ReHFI2[Kd+2];
                            g1 = coeff*ReEZI2[Kd+2];
                          }
                          else {
                            a1 = coeff*ImHFI2[Kd+2]*jK1;
                            g1 = coeff*ImEZI2[Kd+2]*jK1;
                          }
                        }
                        double a2 = 0, g2 = 0;
                        if (abs(Ks)<=2) {
                          const double coeff = jjj(L1,2,L2,K1,-Ks,K2);
                          if (jK1==jK2) {
                            a2 = coeff*ReHFI2[Ks+2];
                            g2 = coeff*ReEZI2[Ks+2];
                          }
                          else {
                            a2 = coeff*ImHFI2[Ks+2]*jK1;
                            g2 = coeff*ImEZI2[Ks+2]*jK1;
                          }
                        }
                        R_EZI2 = g1 + jK2*parityLK2*g2;
                        R_HFI2 = a1 + jK2*parityLK2*a2;
                      }
                      //----------------------------------------------

                      //----------------------------------------------
                      // R(mu=EZI,HFI,NZI;l=0), see Eq. (A42) and (A44)
                      //----------------------------------------------
                      /*
                      double R_EZI0 = 0, R_HFI0 = 0, R_NZI0 = 0;
                      if ((L1==L2) && (jK1==jK2)) {
                        double g1 = 0, a1 = 0, n1 = 0;
                        if (Kd==0) {
                          const double coeff = jjj(L1,0,L2,K1,-Kd,-K2);
                          g1 = coeff*EZI0;
                          a1 = coeff*HFI0;
                          n1 = coeff*NZI0;
                        }
                        double g2 = 0, a2 = 0, n2 = 0;
                        if (Ks==0) {
                          const double coeff = jjj(L1,0,L2,K1,-Ks,K2);
                          g2 = coeff*EZI0;
                          a2 = coeff*HFI0;
                          n2 = coeff*NZI0;
                        }
                        R_EZI0 = g1 + jK2*parityLK2*g2;
                        R_HFI0 = a1 + jK2*parityLK2*a2;
                        R_NZI0 = n1 + jK2*parityLK2*n2;
                      }
                      */
                      //----------------------------------------------

                      // N_K normalization factor, Eq. (A43)
                      double N_K = 1;
                      if (K1==0) N_K /= sqrt(2.0);
                      if (K2==0) N_K /= sqrt(2.0);
                      // Normalisation prefactor in Eq.(A40) and Eq.(A41)
                      const double NormFactor = N_L*N_K*parity(M1+K1);
                      
                      //----------------------------------------------------------------------
                      // Potential-dependent term of diffusion operator, Eq. (A40)
                      //----------------------------------------------------------------------
                      PotDiff = 0;
                      if (Potential) {
                        if ((abs(Ld)<=Lband)&&(parity(Ks)==1)&&
                            (jKd==0)&&(abs(Kd)<=Kband)&&(abs(Ks)<=Kband)) {
                          for (int L=0; L<=Lband; L+=2) {
                            double t1, t2, X;
                            t1 = t2 = 0;
                            X = xlk[L/2,abs(Kd)/2];
                            if ((Kd<=L) && (X!=0))
                              t1 = X * jjj(L1,L,L2,K1,K2-K1,-K2);
                            X = xlk[L/2,Ks/2];
                            if ((Ks<=L) && (X!=0))
                              t2 = parityLK2*jK2*X * jjj(L1,L,L2,K1,-K1-K2,K2);
                            if (t1||t2)
                              PotDiff += (t1+t2) * jjj(L1,L,L2,M1,0,-M1);
                          }
                          PotDiff *= NormFactor;
                        }
                      }
                      //----------------------------------------------------------------------
                      
                      
                      M2max = mini(Mmax,L2);
                      M2min = (diagRC) ? M1 : -M2max;
                      for (M2=M2min;M2<=M2max;M2++) {
                        int Md = M1 - M2;
                        
                        bool diagLKM = diagLK && (jKd==0) && (Md==0);

                        // Pre-compute 3j symbol in Eq. (A41) for l = 2
                        const double Liou3j = (Ld2) ? jjj(L1,2,L2,M1,-Md,-M2) : 0;
                        
                        pS2min = (diagRC) ? pS1 : pSmin;
                        for (pS2=pS2min;pS2<=1;pS2++) {
                          int pSd = pS1 - pS2;
                          qS2max = 1 - abs(pS2);
                          qS2min = (diagRC) ? qS1 : -qS2max;
                          for (qS2=qS2min;qS2<=qS2max;qS2+=2) {
                            bool diagS = (pS1==pS2) && (qS1==qS2);
                            int qSd = qS1 - qS2;
                            pI2min = (diagRC) ? pI1 : -pImax;
                            for (pI2=pI2min;pI2<=pImax;pI2++) {
                              if ((psi==0) && ((pI2+pS2-1)!=M2)) continue;
                              int pId = pI1 - pI2;
                              qI2max = int(2*I) - abs(pI2);
                              qI2min = (diagRC) ? qI1 : -qI2max;
                              for (qI2=qI2min;qI2<=qI2max;qI2+=2) {
                                int qId = qI1 - qI2;
                                bool diagI = (pId==0) && (qId==0);
                                

                                //--------------------------------------
                                // Matrix element of Liouville operator
                                //--------------------------------------
                                // When L1=L2, both the l=0 and l=2 ISTOs must be included   
                                // whereas only l=2 terms appear off the diagonal.
                                //
                                // The isotropic terms are not normalized by any factor.
                                // - The N_K factor is not needed because
                                //   the l=0 ISTO components have not been transformed by
                                //   the K-symmetrization. (following the formula, N_K is
                                //   cancelled by R_0)
                                // - The factor N_L*parityMK1 is canceled by the product of
                                //   wigner3j(L,M;0,0;L,-M) and wigner3j(L,K,0,0,L,-K),
                                //   which therefore do not need to be calculated.
                                int pd = pSd + pId;
                                bool fdiag = diagLKM && (pd==0);
                                double LiouvilleElement = 0;
                                
                                if (Ld2 && (abs(pSd)<=1) && (abs(pId)<=1) && 
                                   ((psi!=0) || (pd==Md)) && (abs(Md)<=2) && 
                                   (abs(pSd)==abs(qSd)) && (abs(pId)==abs(qId))) {
                                  
                                  double d2jjj = d2psi[(pd+2)+(Md+2)*5]*Liou3j;
                                  
                                  // Electronic Zeeman interaction
                                  //-----------------------------------------
                                  if (diagI) { // i.e. pId==0 and qId==0
                                    // Rank-2 term, Eq. (B7)
                                    // Compute Clebsch-Gordan coeffs and S_g Eq. (B8)
                                    double C2, S_g;
                                    if (pSd==0) {
                                      C2 = +sqrt23; // (112|000)
                                      S_g = C2*pS1;
                                    }
                                    else {
                                      C2 = +sqrt12;  // (112|-10-1), (112|101)
                                      S_g = -C2*qSd/sqrt(2.0);
                                    }
                                    LiouvilleElement += NormFactor*d2jjj*R_EZI2*S_g;
                                    // Rank-0 term
                                    if (fdiag) {
                                      const double C0 = -sqrt13; // (110|000)
                                      LiouvilleElement += C0*EZI0*pS1;
                                      //mexPrintf("+EZI0: %e\n",EZI0*C0);
                                    }
                                  }
                                  
                                  // Hyperfine interaction
                                  //-----------------------------------------
                                  if ((I>0) && (pSd*pId==qSd*qId)) {
                                    // Compute Clebsch-Gordan coeffs and S_A Eq. (B7)
                                    double C0, C2, S_A;
                                    if (pId==0) {
                                      if (pSd==0) {
                                        S_A = (pS1*qI1+pI1*qS1)/2.0;
                                        C0 = -sqrt13; // (110|000)
                                        C2 = +sqrt23; // (112|000)
                                      }
                                      else {
                                        S_A = -(pI1*pSd+qI1*qSd)/sqrt(8.0);
                                        C0 = 0; // no rank 0 term
                                        C2 = +sqrt12; // (112|101), (112|-10-1)
                                      }
                                    }
                                    else {
                                      int t = qI1*qId + pI1*pId;
                                      double KI = sqrt(I*(I+1)-t*(t-2.0)/4.0);
                                      if (pSd==0) {
                                        S_A = -(pS1*pId+qS1*qId)*KI/sqrt(8.0);
                                        C0 = 0; // no rank 0 term
                                        C2 = +sqrt12; // (112|011), (112|0-1-1)
                                      }
                                      else {
                                        S_A = pSd*qId*KI/2.0;
                                        C0 = +sqrt13; // (110|1-10), (110|-110)
                                        if (pd==0)
                                          C2 = +sqrt(1.0/6.0); // (112|1-10), (112|-110)
                                        else
                                          C2 = +1.0; // (112|112), (112|-1-1-2)
                                      }
                                    }
                                    // Rank-2 term, Eq. (A41)
                                    LiouvilleElement += NormFactor*d2jjj*R_HFI2 * C2*S_A;
                                    // Rank-0 term
                                    if (fdiag)
                                      LiouvilleElement += HFI0*C0*S_A;
                                  }
                                  
                                  // Nuclear Zeeman interaction
                                  //-----------------------------------------
                                  if (diagS && diagI && fdiag) {
                                    const double C0 = -sqrt13; // (110|000)
                                    LiouvilleElement += pI1*C0*NZI0;
                                  }
                                }
                                //-------------------------------------------
                                
                                
                                //-------------------------------------------
                                // Matrix element of diffusion operator
                                //-------------------------------------------
                                double GammaElement = 0;
                                
                                // Diagonal terms
                                if (diagRC) GammaElement += IsoDiffDiag;
                                
                                // Potential-dependent terms
                                if (Potential)
                                  if (diagS && diagI && (Md==0)&&(jKd==0))
                                    GammaElement += PotDiff;
                                
                                // Exchange term
                                if (Exchange) {
                                  if ((pSd==0) && (pId==0) && diagLKM) {
                                    double t = 0;
                                    if ((qId==0) && (qSd==0)) t += 1.0;
                                    if ((qId==0) && (pS1==0)) t -= 0.5;
                                    if ((pI1==0) && (qSd==0)) t -= 1.0/(2.0*I+1);
                                    GammaElement += t*ExchangeFreq;
                                  }
                                }
                                //-------------------------------------------
                              
                                // Store element values and indices
                                //-------------------------------------------
                                //mexPrintf("%d/%d %d/%d %d/%d %d/%d\n",L1,L2,jK1,jK2,K1,K2,M1,M2);
                                if ((GammaElement!=0) || (LiouvilleElement!=0)) {
                                  if (Display) mexPrintf("     saving element %5d: %f  %f\n",iElement+1,GammaElement,LiouvilleElement);
                                  MatrixRe[iElement] = GammaElement;
                                  MatrixIm[iElement] = -LiouvilleElement;
                                  ridx[iElement] = iRow;
                                  cidx[iElement] = iCol;
                                  iElement++;
                                  if (!diagRC) {
                                    MatrixRe[iElement] = GammaElement;
                                    MatrixIm[iElement] = -LiouvilleElement;
                                    cidx[iElement] = iRow;
                                    ridx[iElement] = iCol;
                                    iElement++;
                                  }
                                  if (iElement>=maxElements)
                                    mexErrMsgTxt("Number of non-zero elements too large");
                                }
                                iCol++;
                                diagRC = false;
                                if (iCol>=maxRows)
                                  mexErrMsgTxt("Dimension too large!");
                                
                              }
                            } 
                          }
                        }
                      }
                    }
                  }
                } // all row index loops

                iRow++;
              }
            }
          }
        }
      }
    }
  }
} // all row index loops

nRows = iRow;
nElements = iElement;

}


//=======================================================================
// Determine basis size
int BasisSize()
//=======================================================================
{

// Basis parameters
const int Lemax = Basis.evenLmax;
const int Lomax = Basis.oddLmax;
const int Kmax = Basis.Kmax;
const int Mmax = Basis.Mmax;
const int jKmin = Basis.jKmin;
const int pSmin = Basis.pSmin;
const int pImax = Basis.pImax;
const int deltaL = Basis.deltaL;
const int deltaK = Basis.deltaK;

// Loop variables
int L1, jK1, K1, M1, pS1, qS1, pI1, qI1;

// Min and max values for loop variables
int K1max, M1max, qS1max, pI1max, qI1max;

int iRow = 0;
double I = Sys.I;

bool NoDirTilt = (Diff.psi==0);

for (L1=0;L1<=Lemax;L1+=deltaL) {
  if (isodd(L1) && (L1>Lomax)) continue;
  for (jK1=jKmin;jK1<=1;jK1+=2) {
    K1max = mini(Kmax,L1);
    for (K1=0;K1<=K1max;K1+=deltaK) {
      if ((K1==0)&(parity(L1)!=jK1)) continue;
      M1max = mini(Mmax,L1);
      for (M1=-M1max;M1<=M1max;M1++) {
        for (pS1=pSmin;pS1<=1;pS1++) {
          qS1max = 1 - abs(pS1);
          for (qS1=-qS1max;qS1<=qS1max;qS1+=2) {
            for (pI1 = -pImax;pI1<=pImax;pI1++) {
              if (NoDirTilt) if ((pI1+pS1-1)!=M1) continue;
              qI1max = int(2*I) - abs(pI1);
              for (qI1=-qI1max;qI1<=qI1max;qI1+=2) {
                //mexPrintf("%3d %3d %3d %3d    %2d %2d %2d %2d\n",L1,jK1,K1,M1,pS1,qS1,pI1,qI1);
                iRow++;
              }
            }
          }
        }
      }
    }
  }
}

int nRows = iRow;
return nRows;

}

//============================================================================
//============================================================================
//============================================================================
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{

  if (nrhs==1) {
    double *jm = mxGetPr(prhs[0]);
    //mexPrintf("%0.10f\n",jjj(jm[0],jm[1],jm[2],jm[3],jm[4],jm[5]));
    return;
  }

  if (nrhs!=3) mexErrMsgTxt("3 input arguments expected!");
  if (nlhs!=5) mexErrMsgTxt("3 output arguments expected!");

  //mexPrintf("Parsing system...\n");
  const mxArray* S = prhs[0];
  Sys.I = mxGetScalar(mxGetField(S,0,"I"));
  Sys.EZI0 = mxGetScalar(mxGetField(S,0,"EZ0"));
  Sys.NZI0 = mxGetScalar(mxGetField(S,0,"NZ0"));
  Sys.HFI0 = mxGetScalar(mxGetField(S,0,"HF0"));

  const mxArray* T;
  T = mxGetField(S,0,"EZ2");
  Sys.ReEZI2 = mxGetPr(T);
  Sys.ImEZI2 = mxGetPi(T);
  T = mxGetField(S,0,"HF2");
  Sys.ReHFI2 = mxGetPr(T);
  Sys.ImHFI2 = mxGetPi(T);
  
  //mexPrintf("Parsing basis structure...\n");
  const mxArray* B = prhs[1];
  Basis.evenLmax = int(mxGetScalar(mxGetField(B,0,"evenLmax")));
  Basis.oddLmax = int(mxGetScalar(mxGetField(B,0,"oddLmax")));
  Basis.Kmax = int(mxGetScalar(mxGetField(B,0,"Kmax")));
  Basis.Mmax = int(mxGetScalar(mxGetField(B,0,"Mmax")));
  Basis.jKmin = int(mxGetScalar(mxGetField(B,0,"jKmin")));
  Basis.pSmin = int(mxGetScalar(mxGetField(B,0,"pSmin")));
  Basis.pImax = int(mxGetScalar(mxGetField(B,0,"pImax")));
  Basis.deltaL = int(mxGetScalar(mxGetField(B,0,"deltaL")));
  Basis.deltaK = int(mxGetScalar(mxGetField(B,0,"deltaK")));
  
  //mexPrintf("Parsing diffusion structure...\n");
  const mxArray* D = prhs[2];
  Diff.Exchange = mxGetScalar(mxGetField(D,0,"Exchange"));
  Diff.psi = mxGetScalar(mxGetField(D,0,"psi"));
  Diff.xlk = mxGetPr(mxGetField(D,0,"xlk"));
  Diff.d2psi = mxGetPr(mxGetField(D,0,"d2psi"));
  Diff.nJumpSites = int(mxGetScalar(mxGetField(D,0,"nJumpSites")));
  Diff.JumpFreq = mxGetScalar(mxGetField(D,0,"JumpFreq"));
  double *R = mxGetPr(mxGetField(D,0,"Diff"));
  Diff.Rxx = R[0];
  Diff.Ryy = R[1];
  Diff.Rzz = R[2];

  int N = BasisSize();

  if (Display) mexPrintf("  basis function count: %d\n",N);
  if (N>maxRows) mexErrMsgTxt("Dimensions too large!");
  
  plhs[0] = mxCreateDoubleMatrix(maxElements,1,mxREAL);
  plhs[1] = mxCreateDoubleMatrix(maxElements,1,mxREAL);
  plhs[2] = mxCreateDoubleMatrix(maxElements,1,mxCOMPLEX);
  ridx = mxGetPr(plhs[0]);
  cidx = mxGetPr(plhs[1]);
  MatrixRe = mxGetPr(plhs[2]);
  MatrixIm = mxGetPi(plhs[2]);
  
  if (Display) mexPrintf("  starting matrix loops...\n");

  makematrix();

  if (Display) mexPrintf("  finishing matrix loops...\n");

  plhs[3] = mxCreateScalarDouble(nRows);
  plhs[4] = mxCreateScalarDouble(nElements);
  
  return;
}
